% initialize_gpsem_mog:
%
% Get initial point for the GPSEM by first running it with a mixture of
% Gaussians model. Hyperparameters for the GPs are obtained by running
% it with imputated latent variables.

function [x_start, lambda_mean, var_y, ...
          kernel_hyper, var_x, x_marg, z_start] = ...
    initialize_gpsem_mog(num_mcmc_samples, K, graph, data, n_types, ...
                         anchors, test_values, covfunc)
                     
% Initial graph pre-processing
num_y = sum(n_types);
y_pos = find(n_types == 1);
num_x = length(n_types) - num_y;
x_pos = zeros(length(n_types), 1);
count_x_pos = 0;
for i = 1:length(n_types)
    if n_types(i) == 0
       count_x_pos = count_x_pos + 1;
       x_pos(i) = count_x_pos;
    end
end
x_parents = cell(num_y, 1);
for y = 1:num_y
    x_parents{y} = x_pos(graph(y_pos(y), :) == 1)';
end

sample_size = size(data, 1);

% Initialize by using MCMC in the mixture of Gaussians model
[test_pred, sampled_x, sampled_lin, sampled_yerr_var, ...
 sampled_xmix_pi, sampled_xmix_mu, sampled_xmix_invcov, sampled_x_z] = ...
          run_mog_mcmc(num_mcmc_samples, K, graph, data, n_types, ...
                       anchors, test_values, []);

% 'x_start' is the initialization for the latent variables                         
x_start = squeeze(sampled_x(num_mcmc_samples, :, :));

% 'x_marg' is the marginal distribution of the exogenous latents
x_marg  = cell(K, 3);
for x = 1:num_x
    x_marg{x} = zeros(K, 3);
end    
for d = 1:num_mcmc_samples
    for s = 1:K
        cov_x = inv(sampled_xmix_invcov{d, s});
        for x = 1:num_x
            x_marg{x}(s, 1) = x_marg{x}(s, 1) + sampled_xmix_pi(d, s);
            x_marg{x}(s, 2) = x_marg{x}(s, 2) + sampled_xmix_mu{d, s}(x);
            x_marg{x}(s, 3) = x_marg{x}(s, 3) + cov_x(x, x);
        end
    end
end

% 'z_start' is the mixture indicator initialization for exogenous 
% latent variables
z_start = zeros(num_x, sample_size);
for x = 1:num_x
    for s = 1:K
        x_marg{x}(s, :) = x_marg{x}(s, :) / num_mcmc_samples;
    end
    for d = 1:sample_size
        z_start(x, d) = sampled_x_z(num_mcmc_samples, d);
    end
end

% Get measurement model estimates as initial values
var_y = mean(sampled_yerr_var)';
lambda_mean = cell(num_y, 1);
for y = 1:num_y
    lambda_mean{y} = mean(sampled_lin{y})';        
end

% Finally, fit GP hyperparameters to imputated latent data
if ~isempty(covfunc)
    var_x = zeros(num_x, 1);          
    kernel_hyper = cell(num_x, 1);
    for x = find(n_types == 0)'
        x_p = x_pos(graph(x, :) == 1); 
        if isempty(x_p)
           var_x(x_pos(x)) = var(x_start(:, x_pos(x)));
           continue
        end
        input_data = x_start(:, x_p);
        output_data = x_start(:, x_pos(x));
        num_params = kernel_nparam(covfunc, length(x_p));
        loghyper = minimize(zeros(num_params, 1), 'gpr', 150, ...
                            covfunc, input_data, output_data);
        kernel_hyper{x_pos(x)} = kernel_gpml_map(covfunc, loghyper, 2);
        kernel_hyper{x_pos(x)}(length(loghyper)) = 1.e-4;
        var_x(x_pos(x)) = exp(loghyper(length(loghyper)))^2 - 1.e-4;
    end
else
    kernel_hyper = [];
    var_x = [];
end
