% mcmc_finite:
%
% This function generates MCMC samples from a linear model with fixed
% basis functions.
%
% IT IS ASSUMED THAT ORDER OF OBSERVED VARIABLES IN THE GRAPH CORRESPONDS
% TO THE ORDER OF COLUMNS IN THE DATASET.
%
% Inputs:
%
% - n_mcmc_samples: number of MCMC samples to be generated.
% - graph: a valid SEM DAG;
% - n_types: a binary vector indicating which variables are observed
%            (== 1) and which are hidden (== 0)
% - basisfunc: vector of strings indicating which set of basis functions is
%              used by each latent variable.
% - lin_var: global prior variance hyperparameter for the linear
%            coefficients
% - hyper_x_var: inverse gamma hyperparameters for conditional latent
%                variance (for endogenous latent variables)
% - hyper_x_mix: parameter for the Dirichlet mixture model (for exogenous 
%                latent variables)
% - hyper_y_var: inverse gamma hyperparameters for measurement variance.
%
% - fixed_hyper_kernel: vector of zeros and ones. Ones indicate which
%                       hyperparameters are fixed using the values at
%                       'hyper_kernel_values'
% - fixed_yerr: vector of zeros and ones. Ones indicate which
%               measurement error variances are fixed using the values at
%               'yerr'
% - fixed_xerr: vector of zeros and ones. Ones indicate which
%               conditional latent variances are fixed using the values at
%               'xerr'
% - fixed_loadings: cell array, one entry per observed variable. Each
%                   entry is a vector indicating which loadings are fixed
%                   to '1' (the size of the vector is the number of parents
%                   of y plus one, where the last entry corresponds to the
%                   intercept).
% - missing_values: a T x 2 matrix indicating missing values to be
%                   predicted. (t, 1) is the row of the t-th missing value
%                   and (t, 2) is the columno of the t-th missing value
% - use_starting: see below.
% - starting_*: initial positions for latent variables, linear
%               coefficients, and error variances. Only used if 
%               'use_starting' is set to 1.
% It is assumed that the prior for the measurement error variances is 
% the constant improper prior.
%
% Prior for kernel hyperparameters will vary according to the kernel.

function [sampled_missing, sampled_x, sampled_f, weights, sampled_lin, ...
          sampled_xerr_var, sampled_xmix, sampled_x_z, sampled_yerr_var, ...
          sampled_yerr_ups] = ...
          mcmc_finite(num_mcmc_samples, graph, n_types, data, lin_var, ...
                      hyper_x_var, hyper_x_mix, hyper_y_var, t_noise, ... 
                      fixed_exo_marg, fixed_exo_marg_values, ...
                      hyper_weight_values, ...
                      fixed_yerr, yerr, fixed_xerr, xerr, ...
                      fixed_loadings, fixed_loading_values, ...
                      missing_values, ...
                      use_starting, ...
                      starting_latents, starting_loadings, starting_yvars, ...
                      starting_weights, starting_xvars, starting_xmix, ...
                      starting_exo_z)

UPS_BOUND = 40; % Prior for degress of freedom 'upsilon' is uniform [0, UPS_BOUND]

BATCH_SIZE = 5000; % Number of samples stored in primary memory
if BATCH_SIZE > num_mcmc_samples
   BATCH_SIZE = num_mcmc_samples;
end

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% PARSE GRAPH information.
%
% x_ychildren: observed children of each latent. It contains the position
%              of each children wrt to the observed variables only, not 
%              the whole graph.
% x_xchildren: latent children of each latent. Again, numbers represent
%              position with respect the latent variables only.
% y_xparents:  latent parents of each observed variable (same indexing
%              comment applies).

num_nodes = size(graph, 1);
latents = find(n_types == 0);  observed = find(n_types == 1);
num_latents = length(latents); num_observed = length(observed);
children = cell(num_nodes, 1); parents = cell(num_nodes, 1);
for v = 1:num_nodes
    children{v} = find(graph(:, v) == 1)';
    parents{v} = find(graph(v, :) == 1);
end

x_ychildren = cell(num_latents, 1); 
x_xchildren = cell(num_latents, 1); x_xparents = cell(num_latents, 1);
exo_latent = zeros(num_latents, 1);
for x = 1:num_latents
    x_pos = latents(x);
    obs_children    = intersect(observed, children{x_pos});
    latent_children = intersect(latents, children{x_pos});
    latent_parents  = intersect(latents, parents{x_pos});
    x_ychildren{x} = zeros(1, length(obs_children));
    for v = 1:length(obs_children)
        x_ychildren{x}(v) = find(observed == obs_children(v), 1);
    end
    x_xchildren{x} = zeros(1, length(latent_children));
    for v = 1:length(latent_children)
        x_xchildren{x}(v) = find(latents == latent_children(v), 1);
    end
    x_xparents{x} = zeros(1, length(latent_parents));
    for v = 1:length(latent_parents)
        x_xparents{x}(v) = find(latents == latent_parents(v), 1);
    end
    exo_latent(x) = isempty(latent_parents);
end
exo_latents = find(exo_latent == 1)'; endo_latents = find(exo_latent == 0)';
y_xparents = cell(1, num_observed);
for y = 1:num_observed
    y_pos = observed(y);
    y_xparents{y} = zeros(1, length(parents{y_pos}));
    for v = 1:length(parents{y_pos})
        y_xparents{y}(v) = find(latents == parents{y_pos}(v), 1);
    end
end

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% ALLOCATE space. Everything is being done in RAM in batches, so this can 
% quickly exhauste the memory. Set BATCH_SIZE accordingly.

N            = size(data, 1);
sampled_x    = zeros(BATCH_SIZE + 1, N, num_latents);
sampled_f    = zeros(BATCH_SIZE + 1, N, num_latents);
sampled_lin  = cell(num_observed, 1);
for v = 1:num_observed
    y = observed(v);
    sampled_lin{v} = zeros(BATCH_SIZE + 1, length(parents{y}) + 1);
end
sampled_xerr_var = zeros(BATCH_SIZE + 1, num_latents);
sampled_xmix     = cell(BATCH_SIZE + 1, length(exo_latents));
sampled_x_z      = zeros(BATCH_SIZE + 1, N, length(exo_latents));
sampled_yerr_var = zeros(BATCH_SIZE + 1, num_observed);
if t_noise
   sampled_yerr_ups = zeros(BATCH_SIZE + 1, num_observed);
else
   sampled_yerr_ups = [];
end

for x = exo_latents
    exo_pos = find(exo_latents == x, 1);
    num_k = length(hyper_x_mix{x, 1});
    for i = 1:(BATCH_SIZE + 1)
        sampled_xmix{i, exo_pos} = zeros(num_k, 3);
    end
end

% Cache all kernel matrices and inverses
weights     = cell(num_latents, 1);
weight_size = zeros(num_latents, 1);
for x = 1:num_latents
    if isempty(x_xparents{x})
       continue
    end
    weight_size(x) = basisfunc(length(x_xparents{x}));
    weights{x} = zeros(BATCH_SIZE + 1, weight_size(x));    
end

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% INITIAL SAMPLING. First, initialize error variances.
% After that, order latents in a partial order where a node 
% cannot come before any of its descendants. Then sample according to
% order.

if use_starting
   sampled_xerr_var(1, :) = starting_xvars;
   sampled_yerr_var(1, :) = starting_yvars;
   if t_noise
      sampled_yerr_ups(1, :) = ones(num_observed, 1) * 40;
   end
   for d = 1:N
      sampled_x(1, d, :) = starting_latents(d, :);
   end
   for y = 1:num_observed
       sampled_lin{y}(1, :) = starting_loadings{y};
   end    
   for x = exo_latents
       exo_pos = find(exo_latents == x, 1);
       if fixed_exo_marg(x)
          sampled_xmix{1, exo_pos} = fixed_exo_marg_values{x};
       else
          sampled_xmix{1, exo_pos} = starting_xmix{x};
       end
       for d = 1:N
           sampled_x_z(1, d, exo_pos) = starting_exo_z(x, d);
       end        
   end
   for x = 1:num_latents
       if isempty(x_xparents{x})
          continue
       end
       % Sample latent functions
       parent_data = basisfunc(squeeze(sampled_x(1, :, :)), x_xparents{x});
       weights{x}(1, :) = starting_weights{x};
       sampled_f(1, :, x) = parent_data * weights{x}(1, :)';
   end
else
    sampled_xerr_var(1, :) = xerr';
    sampled_yerr_var(1, :) = yerr';
    if t_noise
       sampled_yerr_ups(1, :) = ones(num_observed, 1) * 40;
    end
    latent_order = intersect(get_node_order(graph), latents);
    for x_ord = 1:num_latents
        x = latent_order(x_ord);
        if isempty(x_xparents{x})
           sampled_x(1, :, x) = randnorm(N, 1) * ...
                                sqrt(sampled_xerr_var(1, x));
           continue
        end
        % Sample latent functions
        parent_data = basisfunc(squeeze(sampled_x(1, :, :)), x_xparents{x});
        weights{x}(1, :) = randn(1, weight_size(x)) * sqrt(0.1);
        sampled_f(1, :, x) = parent_data * weights{x}(1, :)';
        % Sample latent data points
        sampled_x(1, :, x) = sampled_f(1, :, x) + randnorm(N, 1) * ...
                              sqrt(sampled_xerr_var(1, x));
    end
    for y = 1:num_observed % And now, sample linear coefficients
        n_par = length(y_xparents{y});
        cov_y = eye(n_par + 1);
        sampled_lin{y}(1, :) = randnorm(1, zeros(n_par + 1, 1), [], cov_y)';
    end
    for x = exo_latents
        exo_pos = find(exo_latents == x, 1);
        if fixed_exo_marg(x)
           sampled_xmix{1, exo_pos} = fixed_exo_marg_values{x};
        else
           num_k = length(hyper_x_mix{x, 1});
           sampled_x_z(1, :, exo_pos) = sample(ones(num_k, 1) / num_k, N);
           for k = 1:num_k
               sampled_xmix{1, exo_pos}(k, 1) = 1 / num_k; % Probability 
               sampled_xmix{1, exo_pos}(k, 2) = randn();   % Mean
               sampled_xmix{1, exo_pos}(k, 3) = rand();    % Variance
           end
        end
    end
end

for y = 1:num_observed
    if sum(fixed_loadings{y}) == 0
       continue
    end
    n_par = length(y_xparents{y});
    if ~isempty(fixed_loading_values)
       sampled_lin{y}(1, fixed_loadings{y} == 1) = fixed_loading_values{y}';
    else
       sampled_lin{y}(1, fixed_loadings{y} == 1) = 1;
       if fixed_loadings{y}(n_par + 1)
          sampled_lin{y}(1, n_par + 1) = 0;
       end
    end
end

% Initialize missing values
num_missing = size(missing_values, 1);
count_vars = N * ones(num_observed, 1);
mars_matrix = zeros(size(data));
for t = 1:num_missing
    r = missing_values(t, 1); c = missing_values(t, 2);
    data(r, c) = 0;
    count_vars(c) = count_vars(c) - 1;
end
data_means = sum(data) ./ count_vars';
if num_missing > 0
   sampled_missing = zeros(BATCH_SIZE + 1, num_missing);
else
   sampled_missing = [];
end
for t = 1:num_missing
    r = missing_values(t, 1); c = missing_values(t, 2);
    data(r, c) = data_means(c) + 0.01 * randn();
    sampled_missing(1, t) = data(r, c);
    mars_matrix(r, c) = 1;
end

% Buffers
accept_x = zeros(num_latents, 1);

%disp(sampled_yerr_var(1, :));
if t_noise
   %disp(sampled_yerr_ups(1, :));
end

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% SETUP FILES AND BATCH SIZE INFORMATION

num_batches = ceil(num_mcmc_samples / BATCH_SIZE);
num_batch_samples = BATCH_SIZE * ones(num_batches, 1);
num_batch_samples(num_batches) = num_mcmc_samples - BATCH_SIZE * (num_batches - 1);

file_x      = fopen('sample_x.txt','w');
file_y      = fopen('sample_y.txt','w');
file_lambda = fopen('sample_lambda.txt','w');
file_yvar   = fopen('sample_yvar.txt','w');
file_exo    = fopen('sample_exo.txt', 'w');
file_xvar   = fopen('sample_xvar.txt','w');
file_w      = fopen('sample_w.txt','w');
file_z      = fopen('sample_z.txt','w');
file_yups   = fopen('sample_yups.txt','w'); 

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% PERFORM SAMPLING

total_iter = 0;

for bt = 1:num_batches
for iter = 2:(num_batch_samples(bt) + 1)

    total_iter = total_iter + 1; disp(sprintf('Iteration %d', total_iter));

    sampled_x(iter, :, :) = sampled_x(iter - 1, :, :);
    sampled_xerr_var(iter, :) = sampled_xerr_var(iter - 1, :);
    for x = 1:num_latents
        if ~isempty(x_xparents{x})
           weights{x}(iter, :) = weights{x}(iter - 1, :);
           sampled_f(iter, :, x) = sampled_f(iter - 1, :, x);
        end
    end
    for y = 1:num_observed
        sampled_yerr_var(iter, y) = sampled_yerr_var(iter - 1, y);
        sampled_lin{y}(iter, :) = sampled_lin{y}(iter - 1, :);
    end
    if t_noise
        for y = 1:num_observed
            sampled_yerr_ups(iter, y) = sampled_yerr_ups(iter - 1, y);
        end
    end
            
    %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
    % Sample latent variable conditional variance hyperparameters
    % (for endogenous latent variables)
    
    for x = endo_latents
        if fixed_xerr(x)
           sampled_xerr_var(iter, x) = sampled_xerr_var(iter - 1, x);
           continue
        end
        ss_x = sum((sampled_x(iter - 1, :, x) - sampled_f(iter, :, x)) .^ 2);
        ga = hyper_x_var(1) + N / 2;
        gb = hyper_x_var(2) + ss_x / 2;
        sampled_xerr_var(iter, x) = gb / randgamma(ga);        
    end

    %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
    % Sample latent variable marginal (mixture of Gaussians) parameters
    % (for exogenous latent variables)
    
    for x = exo_latents        
        exo_pos = find(exo_latents == x, 1);
        num_k = length(hyper_x_mix{x, 1});
        % First, sample means and variances for each mixture component        
        if fixed_exo_marg(x)
           sampled_xmix{iter, exo_pos} = fixed_exo_marg_values{x};
        else
           for k = 1:num_k
               k_rows = find(sampled_x_z(iter - 1, :, exo_pos) == k);
               s_x = sum(sampled_x(iter - 1, k_rows, x));            
               % Sample mean give variance
               v_mean = 1 / (1 / hyper_x_mix{x, 2} + ...
                        length(k_rows) / sampled_xmix{iter - 1, exo_pos}(k, 3));
               m_mean = v_mean * s_x / sampled_xmix{iter - 1, exo_pos}(k, 3);
               new_mean = randn() * sqrt(v_mean) + m_mean;
               sampled_xmix{iter, exo_pos}(k, 2) = new_mean;
               % Sample variance given mean
               ss_x = sum((sampled_x(iter - 1, k_rows, x) - new_mean).^2);
               ga = hyper_x_mix{x, 3}(1) + length(k_rows) / 2;
               gb = hyper_x_mix{x, 3}(2) + ss_x / 2;
               sampled_xmix{iter, exo_pos}(k, 3) = gb / randgamma(ga);
           end
        end
        % Now sample mixture component indicators given x and means
        for d = 1:N
            weights_x = zeros(num_k, 1);
            for k = 1:num_k
                m = sampled_xmix{iter, exo_pos}(k, 2); 
                v = sampled_xmix{iter, exo_pos}(k, 3);
                xd = sampled_x(iter - 1, d, x);
                weights_x(k) = -0.5 * log(v) - 0.5 * (xd - m)^2 / v + ...
                               log(sampled_xmix{iter - 1, exo_pos}(k, 1));
            end
            weights_x = weights_x - max(weights_x); weights_x = exp(weights_x);
            sampled_x_z(iter, d, exo_pos) = sample(weights_x/sum(weights_x));
        end
        % Finally, sample mixture distribution
        alpha = hyper_x_mix{x, 1};
        for k = 1:num_k
            alpha(k) = alpha(k) + sum(sampled_x_z(iter, :, exo_pos) == k);
        end
        new_prob = randdirichlet(alpha);
        for k = 1:num_k
            sampled_xmix{iter, exo_pos}(k, 1) = new_prob(k);
        end        
    end
    
    %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
    % Sample measurement error hyperparameters
    
    if t_noise
        %TODO SKIP MISSING DATA
       for y = 1:num_observed
           if fixed_yerr(y)
              continue
           end
           old_s2  = sampled_yerr_var(iter, y);
           old_ups = sampled_yerr_ups(iter, y);
           [new_s2 new_ups logfwd logbkw] = t_noise_proposal(old_s2, old_ups);
           if isinf(logfwd)
              continue;
           end
           if length(y_xparents{y}) > 1
              x_data = [squeeze(sampled_x(iter - 1, :, y_xparents{y})) repmat(1, N, 1)];
           else
              x_data = [sampled_x(iter - 1, :, y_xparents{y})' repmat(1, N, 1)];
           end
           ss_ys = (data(:, y) - x_data * sampled_lin{y}(iter, :)') .^ 2;
           logfactor_old = -0.5 * (old_ups + 1) * sum(log(1 + ss_ys ./ (old_ups * old_s2))) + ...
                           N * (gammaln(0.5 * (old_ups + 1)) - gammaln(0.5 * old_ups) -...
                           0.5 * log(old_ups) -0.5 * log(old_s2));
           logfactor_new = -0.5 * (new_ups + 1) * sum(log(1 + ss_ys ./ (new_ups * new_s2))) + ...
                           N * (gammaln(0.5 * (new_ups + 1)) - gammaln(0.5 * new_ups) -...
                           0.5 * log(new_ups) - 0.5 * log(new_s2));
           logfactor_old = logfactor_old + loginvgammapdf(old_s2, hyper_y_var(1), hyper_y_var(2));
           logfactor_new = logfactor_new + loginvgammapdf(new_s2, hyper_y_var(1), hyper_y_var(2));
           ratio         = exp(logfactor_new + logbkw - logfactor_old - logfwd);
           if rand() < ratio
              sampled_yerr_var(iter, y) = new_s2;
              sampled_yerr_ups(iter, y) = new_ups;
           end
           %disp(sampled_yerr_var(iter, :));
           %disp(sampled_yerr_ups(iter, :));
       end
    else
        %TODO SKIP MISSING DATA
        for y = 1:num_observed
            if fixed_yerr(y)
               sampled_yerr_var(iter, y) = sampled_yerr_var(iter - 1, y);
               continue
            end
            if length(y_xparents{y}) > 1
               x_data = [squeeze(sampled_x(iter - 1, :, y_xparents{y})) repmat(1, N, 1)];
            else
               x_data = [sampled_x(iter - 1, :, y_xparents{y})' repmat(1, N, 1)];
            end
            ss_y = sum((data(:, y) - x_data * sampled_lin{y}(iter - 1, :)') .^ 2);
            ga = hyper_y_var(1) + N / 2;
            gb = hyper_y_var(2) + ss_y / 2;
            sampled_yerr_var(iter, y) = gb / randgamma(ga);
        end    
        %disp(sampled_yerr_var(iter, :));
    end
    
    %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
    % Sample latent function weights given latent variables and data
    
    sampled_x(iter, :, :) = sampled_x(iter - 1, :, :);
    for x = 1:num_latents
        if ~isempty(parents{x})
            parent_data = basisfunc(squeeze(sampled_x(iter, :, :)), x_xparents{x});
            V = parent_data' * parent_data / sampled_xerr_var(iter, x) + ...
                eye(weight_size(x)) / hyper_weight_values(x);
            Sigma = inv(V);
            mu = Sigma * sum(repmat(sampled_x(iter, :, x)', 1, weight_size(x)) .* parent_data)' / sampled_xerr_var(iter, x);
            weights{x}(iter, :) = randnorm(1, mu, [], Sigma)';
            sampled_f(iter, :, x) = parent_data * weights{x}(iter, :)';
        end
    end
    
    %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
    % Hard part: sample latent variables given latent functions, 
    % linear coefficients and data    
    
    for x = 1:num_latents 
        % Collect factors for current latent variable
        logfactors1 = latent_logfactors(x);
        % Collect factors for proposed latent variable
        old_xv = sampled_x(iter, :, x);
        new_xvs = randn(N, 1) * sqrt(1) + sampled_x(iter, :, x)';
        sampled_x(iter, :, x) = new_xvs;
        logfactors2 = latent_logfactors(x);
        % Now, decide Metropolis-Hastings move
        rn = rand(N, 1);
        ratios = exp(logfactors2 - logfactors1);
        rejects = rn > ratios;
        accepts = setdiff(1:N, rejects);
        if sum(rejects) > 0 
           % Reject these
           sampled_x(iter, rejects, x) = old_xv(rejects); 
        end
        if ~isempty(accepts)
           % Accept these
           update_x_functions(x, accepts);
           accept_x(x) = accept_x(x) + length(accepts);
        end
    end

    %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
    % Sample linear coefficients given latent variables and data
    
    if t_noise
        for y = 1:num_observed        
            recN  = mars_matrix(:, y) == 0;
            n_par = length(y_xparents{y});
            for yp = 1:n_par
                if fixed_loadings{y}(yp)
                   continue
                end
                old_l = sampled_lin{y}(iter, yp);
                logfactor_old = sum(t_logfactors(y, recN)) - 0.5 * old_l^2 / lin_var;
                new_l = sampled_lin{y}(iter, yp) + sqrt(0.1) * randn();
                sampled_lin{y}(iter, yp) = new_l;
                logfactor_new = sum(t_logfactors(y, recN)) - 0.5 * new_l^2 / lin_var;
                ratio = exp(logfactor_new - logfactor_old);
                if rand() > ratio % Reject
                   sampled_lin{y}(iter, yp) = old_l;    
                end
            end
        end
    else
        for y = 1:num_observed        
            recN  = mars_matrix(:, y) == 0;
            n_par = length(y_xparents{y});
            if length(y_xparents{y}) > 1
               x_data = [squeeze(sampled_x(iter, recN, y_xparents{y})) repmat(1, sum(recN), 1)];
            else
               x_data = [sampled_x(iter, recN, y_xparents{y})' repmat(1, sum(recN), 1)];
            end
            cov_y = x_data' * x_data / sampled_yerr_var(iter, y);
            for i = 1:(n_par + 1)
                cov_y(i, i) = cov_y(i, i) + 1 / lin_var;
            end
            cov_y = inv(cov_y);
            mean_y = cov_y * sum(repmat(data(recN, y), 1, n_par + 1) .* x_data)' / ...
                             sampled_yerr_var(iter, y); 
            if sum(fixed_loadings{y} == 0)
               sampled_lin{y}(iter, :) = randnorm(1, mean_y, [], cov_y)';        
            else
               fr = fixed_loadings{y} == 0; % Free coefficients
               fx = fixed_loadings{y} == 1; % Fixed coefficients
               if ~isempty(fixed_loading_values)
                  sampled_lin{y}(iter, fx) = fixed_loading_values{y}';
               else
                  sampled_lin{y}(iter, fx) = 1;            
                  if fixed_loadings{y}(n_par + 1)
                     sampled_lin{y}(iter, n_par + 1) = 0;
                  end
               end            
               reg_coeff = cov_y(fr, fx) * inv(cov_y(fx, fx)); % IS THIS CORRECT? TODO
               mean_f = reg_coeff * (sampled_lin{y}(iter, fx)' - mean_y(fx));
               cov_f  = cov_y(fr, fr) - reg_coeff * cov_y(fx, fr);
               sampled_lin{y}(iter, fr) = randnorm(1, mean_f, [], cov_f)';
             end
        end
    end
    
    %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
    % Sample missing values
    
    if num_missing > 0
       sds = sqrt(sampled_yerr_var(iter, :));
       rand_pool = randn(num_missing, 1);
       for t = 1:num_missing
           d = missing_values(t, 1);
           y = missing_values(t, 2);        
           m_t = sampled_lin{y}(iter, :) * [squeeze(sampled_x(iter, d, y_xparents{y})); 1];
           data(d, y) = m_t + sds(y) * rand_pool(t);
           sampled_missing(iter, t) = m_t; % Report mean only
       end
    end
end % batch

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%% NEXT BATCH preparation

last_s = num_batch_samples(bt) + 1;
print_to_file(last_s); % Print whole batch to file
sampled_x(1, :, :) = sampled_x(last_s, :, :);
sampled_f(1, :, :) = sampled_f(last_s, :, :);
for v = 1:num_observed
    sampled_lin{v}(1, :) = sampled_lin{v}(last_s, :);
end
sampled_xerr_var(1, :) = sampled_xerr_var(last_s, :);
for x = exo_latents
    exo_pos = find(exo_latents == x, 1);
    sampled_xmix{1, exo_pos} = sampled_xmix{last_s, exo_pos};
end
sampled_x_z(1, :, :)   = sampled_x_z(last_s, :, :);
sampled_yerr_var(1, :) = sampled_yerr_var(last_s, :);
for x = 1:num_latents
    if ~isempty(parents{x})
       weights{x}(1, :) = weights{x}(last_s, :);
    end
end
if num_missing > 0
   sampled_missing(1, :) = sampled_missing(last_s, :);
end

end % all batches

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%% Local functions

function logfactors = latent_logfactors(x)
    % Factor due to latent parent function/marginal
    if ~exo_latent(x)
       logfactors = (-0.5 * (sampled_x(iter, :, x) - sampled_f(iter, :, x)).^2 ./ ...
                           sampled_xerr_var(iter, x))';
    else
       exo_pos = find(exo_latents == x, 1);
       z_x = sampled_x_z(iter, :, exo_pos);
       num_k = length(hyper_x_mix{x, 1});
       mms = zeros(N, 1); vrs = zeros(N, 1);
       for zz = 1:num_k
           z_idx = find(z_x == zz);
           mms(z_idx) = sampled_xmix{iter, exo_pos}(zz, 2);
           vrs(z_idx) = sampled_xmix{iter, exo_pos}(zz, 3);
       end
       logfactors = -0.5 * (sampled_x(iter, :, x)' - mms).^2 ./ vrs;
    end
    % Factors due to observed (and recorded) children    
    if t_noise
       for y_c = x_ychildren{x}
           recN  = (mars_matrix(:, y_c) == 0);
           logfactors(recN) = logfactors(recN) + t_logfactors(y_c, recN);
       end
    else
       for y_c = x_ychildren{x}
           recN  = (mars_matrix(:, y_c) == 0);
           if length(y_xparents{y_c}) > 1
              x_data = [squeeze(sampled_x(iter, recN, y_xparents{y_c})) repmat(1, sum(recN), 1)];
           else
              x_data = [sampled_x(iter, recN, y_xparents{y_c})' repmat(1, sum(recN), 1)];
           end
           b_y = sampled_lin{y_c}(iter, :)';
           logfactors(recN) = logfactors(recN) - ...
                        0.5 * (data(recN, y_c) - x_data * b_y).^2 ./ sampled_yerr_var(iter, y_c);
       end
    end
    % Factors due to latent children functions
    for x_c = x_xchildren{x}
        parent_data = basisfunc(squeeze(sampled_x(iter, :, :)), x_xparents{x_c});
        m_f = weights{x_c}(iter, :) * parent_data';
        logfactors = logfactors - ...
                       (0.5 * (sampled_x(iter, :, x_c) - m_f).^2 ./ sampled_xerr_var(iter, x_c))';
    end            
end

function update_x_functions(x, D)
    for x_c = x_xchildren{x}
        parent_data = basisfunc(squeeze(sampled_x(iter, D, :)), x_xparents{x_c});
        sampled_f(iter, D, x_c) = weights{x_c}(iter, :) * parent_data';
    end            
end

function [new_s2 new_ups logfwd logbkw] = t_noise_proposal(s2, upsilon)
    ratio     = 3/4;

    if upsilon <= 0 || upsilon > UPS_BOUND || s2 <= 0
       new_s2 = 0; new_ups = 0; logfwd = -Inf; logbkw = -Inf;
       return
    end
    
    window    = s2 / ratio - s2 * ratio;
    new_s2    = s2 * ratio + rand() * window;
    logfwd    = log(1 / window);
    window_bk = new_s2 / ratio - new_s2 * ratio;
    logbkw    = log(1 / window_bk); 

    top_window = upsilon / ratio;
    if top_window > UPS_BOUND
       top_window = UPS_BOUND;
    end        
    window    = top_window - upsilon * ratio;
    new_ups   = upsilon * ratio + rand() * window;
    logfwd    = logfwd + log(1 / window);
    top_b_window = new_ups / ratio;
    if top_b_window > UPS_BOUND
       top_b_window = UPS_BOUND;
    end            
    window_bk = top_b_window - new_ups * ratio;
    logbkw    = logbkw + log(1 / window_bk); 
end

function d = loginvgammapdf(value, alpha, beta)
   d = -(alpha + 1) * log(value) - beta / value  + ...
       alpha * log(beta) - gammaln(alpha);
end

function f = t_logfactors(y, D)
   ups   = sampled_yerr_ups(iter, y);
   s2    = sampled_yerr_var(iter, y);
   sizeD = sum(D);
   if length(y_xparents{y}) > 1
      x_data = [squeeze(sampled_x(iter, D, y_xparents{y})) repmat(1, sizeD, 1)];
   else
      x_data = [sampled_x(iter, D, y_xparents{y})' repmat(1, sizeD, 1)];
   end
   ss_ys = (data(D, y) - x_data * sampled_lin{y}(iter, :)') .^ 2;

   f = -0.5 * (ups + 1) * log(1 + ss_ys ./ (ups * s2));
end

% Print all points in the batch to the respective files

function print_to_file(endpoint)
   disp('Printing to file...');
   % Dump latent variables
   x_format = ' %.3f';
   x_format_r = x_format;
   for w = 1:(num_latents - 1)
       x_format_r = strcat(x_format_r, x_format);
   end
   x_format_r = strcat(x_format_r, '\n');
   for s = 2:endpoint
       fprintf(file_x, x_format_r, sampled_x(s, :, :));    
   end
   % Dump imputed data
   if num_missing > 0
      y_format = ' %.3f';
      y_format_r = y_format;
      for w = 1:(num_missing - 1)
          y_format_r = strcat(y_format_r, y_format);
      end
      y_format_r = strcat(y_format_r, '\n');
      for s = 2:endpoint
          fprintf(file_y, y_format_r, sampled_missing(s, :));    
      end
   end
   % Dump latent variances
   xvar_format = ' %.3f';
   for s = 2:endpoint
       for x = 1:num_latents
           fprintf(file_xvar, xvar_format, sampled_xerr_var(s, x));    
       end
       fprintf(file_xvar, '\n');    
   end   
   % Dump exogenous information
   exo_format = ' %.3f';
   for s = 2:endpoint
       for x = 1:num_latents
           if ~isempty(parents{x})
              continue
           end
           exo_pos = find(exo_latents == x, 1);
           num_k = length(hyper_x_mix{x, 1});
           for zz = 1:num_k
               fprintf(file_exo, exo_format, sampled_xmix{s, exo_pos}(zz, 1)); 
               fprintf(file_exo, exo_format, sampled_xmix{s, exo_pos}(zz, 2)); 
               fprintf(file_exo, exo_format, sampled_xmix{s, exo_pos}(zz, 3)); 
           end
       end
       fprintf(file_exo, '\n');    
   end   
   % Dump linear coefficient hyperparameters
   lambda_format = ' %.3f';
   for s = 2:endpoint
       for y = 1:num_observed
           lambda_format_r = lambda_format;
           for w = 1:length(y_xparents{y})
               lambda_format_r = strcat(lambda_format_r, lambda_format);
           end
           fprintf(file_lambda, lambda_format_r, sampled_lin{y}(s, :));    
       end
       fprintf(file_lambda, '\n');    
   end   
   % Dump variances
   var_format = ' %.3f';
   for s = 2:endpoint
       for y = 1:num_observed
           fprintf(file_yvar, var_format, sampled_yerr_var(s, y));    
       end
       fprintf(file_yvar, '\n');
   end   
   % Dump latent function parameters
   w_format = ' %.3f';
   for s = 2:endpoint
       for x = 1:num_latents
           if isempty(parents{x})
              continue
           end           
           w_format_r = w_format;
           for w = 1:length(weight_size(x))
               w_format_r = strcat(w_format_r, w_format);
           end
           fprintf(file_w, w_format, weights{x}(s, :));    
       end
       fprintf(file_w, '\n');
   end
   % Dump t degress of freedom
   if t_noise
      for s = 2:endpoint
          for y = 1:num_observed
              fprintf(file_yups, var_format, sampled_yerr_ups(s, y));    
          end
          fprintf(file_yups, '\n');    
      end   
   end
end

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% CLEAN-UP

fclose(file_x); fclose(file_y); fclose(file_lambda); fclose(file_yvar); fclose(file_yups);
fclose(file_z); fclose(file_exo); fclose(file_xvar); fclose(file_w); 

accept_x = accept_x / (num_mcmc_samples * N);
disp(accept_x);

end

