%%%%%%%%%%%%
% pred_llik_full:
%
% Compute the loglikelihood of each individual test point.
%
% Compute the loglikelihood of each individual test point.
%
% Notice that we treat the test points as independent, and report the
% loglikelihood individually.
%
% Inputs:
%
% - graph: the graphical structure;
% - K: number of mixture components;
% - num_mcmc: number of samples to take into account in the calculation
%             (might be less than the one in the actual log files);
% - num_latents: number of latent variables in the graph;
% - test_rows: which rows of the 'data' matrix are the test points?
% - data: measurements of the observable variables;
% - dir_name: directory where all the logfiles are;
% - burn_in: number of samples to skip before calculating predictive
%            loglikelihood;
% - subsample: if 1, it knows it should used only a subset of the training
%              points
% - subsample_size: size of the subsample, if subsample is 1
% - NUM_FOLDS: number of cross-validation folds;
%
% Outputs:
%
% - llik: cell structure with NUM_FOLDER entries. In each cell, a vector with
%         as many entries as the test set. For each entry, the MCMC estimate
%         of the predictive loglikelihood of that respective test point;
% - y_fake: posterior samples of the observed variables for the last
%           training folder;
% - x_fake: posterior samples of the latents variables for the last
%           training folder;
%
% IT IS ASSUMED THAT ALL THE LOG FILES HAVE DEFAULT NAMES AS GIVEN BY 
% 'TEST_GPSEM_MCMC_FULL'.

function [llik y_fake x_fake] = ...
    pred_llik_full(graph, K, num_mcmc, covfunc, num_latents, test_rows, ...
                   data, dir_name, burn_in, subsample, subsample_size, NUM_FOLDS)

% convert to cell if needed               
if ischar(covfunc), covfunc = cellstr(covfunc); end 

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%% INITIALIZATION

%%% Get some graph information
num_y        = size(graph, 1) - num_latents;
exo_x        = get_exo_vars(graph(1:num_latents, 1:num_latents));
endo_x       = setdiff(1:num_latents, exo_x);
num_exo      = length(exo_x);
latent_order = get_node_order(graph(1:num_latents, 1:num_latents));
latent_order = setdiff(latent_order, exo_x)';
x_parents    = graph((num_latents + 1):size(graph, 1), 1:num_latents);
num_parents  = sum(x_parents, 2);
x_xparents   = cell(num_latents, 1);
for i = 1:num_latents
    x_xparents{i} = find(graph(i, :) == 1);
end
num_hyper = zeros(num_latents, 1);
for x = endo_x
    num_hyper(x) = kernel_nparam(covfunc, length(x_xparents{x}));
end
total_hyper = sum(num_hyper);
xp_list = cell(num_y, 1);
for y = 1:num_y
    xp_list{y} = [find(x_parents(y, :) == 1) (num_latents + 1)];
end
num_lambda  = sum(sum(x_parents)) + num_y; % Total number of parents, plus intercepts

%%% Allocate some buffer information
lambda      = zeros(num_y, num_latents + 1);
llik        = cell(NUM_FOLDS, 1);
y_fake      = zeros(num_mcmc - burn_in, num_y);
x_fake      = zeros(num_mcmc - burn_in, num_latents);

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%% ITERATION

%%% Now, iterate through folds, and within folds iterate through MCMC
%%% samples
for k = 1:NUM_FOLDS
    
    %%% Get some data information
    test_size  = length(test_rows{k});
    if ~subsample
       train_size = size(data{k}, 1) - length(test_rows{k});
    else
       train_size = subsample_size;
    end

    %%% Allocate more buffer information
    mu_y        = zeros(test_size, num_y);
    buffer_xp   = zeros(num_latents * train_size, 1);
    llik{k}     = zeros(test_size, 1);
    llik_k      = zeros(num_mcmc - burn_in, test_size);
    sampled_hgp = cell(num_latents, 1);
    
    %%% Open files

    file_id_xp  = fopen(strcat(dir_name, '\sample_x.', int2str(k), '.log'), 'r');
    file_id_fp  = fopen(strcat(dir_name, '\sample_f.', int2str(k), '.log'), 'r');
    file_id_fh  = fopen(strcat(dir_name, '\sample_hgp.', int2str(k), '.log'), 'r');
    file_id_xv  = fopen(strcat(dir_name, '\sample_xvar.', int2str(k), '.log'), 'r');    
    file_id_exo = fopen(strcat(dir_name, '\sample_exo.', int2str(k), '.log'), 'r');
    file_id_l   = fopen(strcat(dir_name, '\sample_lambda.', int2str(k), '.log'), 'r');
    file_id_v   = fopen(strcat(dir_name, '\sample_yvar.', int2str(k), '.log'), 'r');
    
    xpdata_line = textscan(file_id_xp,  '%f', num_latents);
    fpdata_line = textscan(file_id_fp,  '%f', num_latents);
    fhdata_line = textscan(file_id_fh,  '%f', total_hyper);
    varx_line   = textscan(file_id_xv,  '%f', num_latents);
    exopar_line = textscan(file_id_exo, '%f', num_exo * K * 3);
    lambda_line = textscan(file_id_l,   '%f', num_lambda);
    var_line    = textscan(file_id_v,   '%f', num_y);

    %%% Iterate through the MCMC samples
    mcmc_count  = 1;    
    while ~feof(file_id_v) && mcmc_count <= num_mcmc 
        disp(sprintf('folder %d, mcmc #%d', k, mcmc_count));
        
        %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
        % START Read sampled training latent variables, and parameters
        
        %----- Read latent variables
        buffer_pos = 1;
        for d = 1:train_size
            buffer_xp(buffer_pos:(buffer_pos + num_latents - 1)) = xpdata_line{1};
            buffer_pos = buffer_pos + num_latents;
            xpdata_line = textscan(file_id_xp, '%f', num_latents);
        end
        sampled_xp = reshape(buffer_xp, train_size, num_latents); % Notice unusual representation        
        %----- Read latent variables
        buffer_pos = 1;
        for d = 1:train_size
            buffer_xp(buffer_pos:(buffer_pos + num_latents - 1)) = fpdata_line{1};
            buffer_pos = buffer_pos + num_latents;
            fpdata_line = textscan(file_id_fp, '%f', num_latents);
        end
        sampled_fp = reshape(buffer_xp, train_size, num_latents); % Notice unusual representation        
        %----- Read Gaussian process hyperparameters        
        buffer_pos = 1;
        buffer_hgp = fhdata_line{1};
        for x = endo_x
            sampled_hgp{x} = buffer_hgp(buffer_pos:(buffer_pos + num_hyper(x) - 1));
            for j = 1:length(sampled_hgp{x})
                if sampled_hgp{x}(j) == 0
                   sampled_hgp{x}(j) = 1.e-4; %FIX A ROUNDING MISTAKE WHEN FILES WERE PRINTED
                end
            end
            buffer_pos = buffer_pos + num_hyper(x);
        end
        fhdata_line = textscan(file_id_fh,  '%f', total_hyper);
        %----- Read latent conditional variances
        vx = varx_line{1}';
        varx_line   = textscan(file_id_xv,  '%f', num_latents);
        %----- Read exogenenous mixture parameters
        exo_params = reshape(exopar_line{1}, K * 3, num_exo)';
        exopar_line = textscan(file_id_exo, '%f', num_exo * K * 3);
        %----- Read measurement model
        p_count = 1;
        for y = 1:num_y
            lambda(y, xp_list{y}) = lambda_line{1}(p_count:(p_count + num_parents(y)))';
            p_count = p_count + 1 + num_parents(y);
        end 
        lambda_line = textscan(file_id_l,   '%f', num_lambda);
        v = var_line{1}';
        var_line    = textscan(file_id_v,   '%f', num_y);
        
        % END Read sampled training latent variables, and parameters
        %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
        
        %%% Decides whether to count this or not
        if mcmc_count <= burn_in
           mcmc_count = mcmc_count + 1;
           continue;
        end
        
        %%% Now, sample new latent variables
        xstar = zeros(test_size, num_latents + 1); 
        % Exogenous variables first
        for x_pos = 1:num_exo
            prob = zeros(K, 1);
            for kk = 1:K
                prob(kk) = exo_params(x_pos, (kk - 1) * 3 + 1);
            end
            z = sample(prob, test_size)';
            mean_x = exo_params(x_pos, (z - 1) * 3 + 2)';
            var_x  = exo_params(x_pos, (z - 1) * 3 + 3)';
            xstar(:, exo_x(x_pos)) = randn(test_size, 1) .* sqrt(var_x) + mean_x;
        end
        % Endogenous variables (following partial order)
        for x = latent_order
            logtheta = kernel_gpml_map(covfunc, sampled_hgp{x}, 1);
            in_data  = sampled_xp(:, x_xparents{x});
            K_M      = feval(covfunc{:}, logtheta, in_data);
            L        = chol(K_M);
            alpha    = solve_chol(L, sampled_fp(:, x));
            
            [kk_star k_star] = feval(covfunc{:}, logtheta, in_data, xstar(:, x_xparents{x}));
            f_star   = k_star' * alpha;
            v_alpha  = L' \ k_star;
            v_star   = kk_star - sum(v_alpha .* v_alpha, 1)'; %kk_star - diag(v_alpha' * v_alpha)
            
            xstar(:, x) = randn(test_size, 1) .* sqrt(v_star + vx(x)) + f_star;
        end
        % Add intercept term
        xstar(:, num_latents + 1) = ones(test_size, 1);        
        
        %%% Calculate the loglikelihood of Y* given X* and store it
        for d = 1:test_size
            mu_y(d, :) = sum(lambda .* repmat(xstar(d, :), num_y, 1), 2)';
        end
        v_matrix = repmat(v, test_size, 1);
        llik_k(mcmc_count - burn_in, :) = sum(-0.5 * log(v_matrix) ...
                                          -0.5 * ((data{k}(test_rows{k}, :) - mu_y) .^ 2) ./ v_matrix, 2);
                                 
        %%% Fake data
        if k == NUM_FOLDS
           y_fake(mcmc_count - burn_in, :) = randn(1, num_y) .* sqrt(v) + mu_y(1, :);
           x_fake(mcmc_count - burn_in, :) = xstar(1, 1:num_latents);
        end
        
        %%% End of MCMC iteration                              
        mcmc_count = mcmc_count + 1;
    end
    
    %%% End of folder iteration
    
    llik_k = llik_k(1:(mcmc_count - burn_in - 1), :);
    num_llik_k = size(llik_k, 1);    
    for d = 1:test_size
        m = max(llik_k(:, d));
        llik{k}(d) = log(sum(exp(llik_k(:, d) - m))) + m...
                     -0.5 * num_y * log(2 * pi) -log(num_llik_k);        
    end
    
    fclose(file_id_xp);  fclose(file_id_fp);  fclose(file_id_fh); fclose(file_id_xv);
    fclose(file_id_exo); fclose(file_id_l);  fclose(file_id_v);

end



% Example code:
%
% all_data = cell(NUM_FOLDS, 1); for i = 1:NUM_FOLDS, all_data{i} = normalize_training(raw_data, all_train_points{i}, zeros(size(raw_data))); end
% burn_in = 1000; dir_name = '.'; covfunc = {'covSum', {'covSEiso','covNoise'}};  num_mcmc_samples = 6000;
% subsample = 1; subsample_size = 300;
% [llik y_fake x_fake] = pred_llik_full(graph, num_x_mix, num_mcmc_samples, covfunc, num_latents, all_test_points, all_data, dir_name, burn_in, subsample, subsample_size, NUM_FOLDS);
